from dataclasses import dataclass
from difflib import SequenceMatcher, unified_diff
from pathlib import Path

from rich.markdown import Markdown

from src.constants import console

EDIT_FILE_TOOL = {
    "type": "function",
    "function": {
        "name": "edit_file",
        "description": "Edits (or creates) a file. Each edit must specify **whole lines**, including any leading spaces or tabs, as the `find` string; the tool does **not** perform in-line substring replacements. It searches for a contiguous block of lines that exactly matches the provided `find` text (including newline characters). If a matching block is found, those lines are removed and replaced by the `replace` string, which also must contain the complete lines with correct indentation. To modify part of a line, include the entire original line in `find` and provide the fully updated line in `replace`. If you want to create a file or replace the entire contents of an existing file, only return a single edit where 'find' is empty and 'replace' is the new entire contents of the file. You must use the read_file tool before editing a file.",
        "parameters": {
            "type": "object",
            "properties": {
                "path": {
                    "type": "string",
                    "description": "Relative path of the file",
                },
                "edits": {
                    "type": "array",
                    "description": "List of edits to apply.",
                    "items": {
                        "type": "object",
                        "properties": {
                            "find": {
                                "type": "string",
                                "description": "Content to find. Include additional previous/following lines if necessary to uniquely identify the section.",
                            },
                            "replace": {
                                "type": "string",
                                "description": "The content to replace with. Must have the correct spacing and indentation for all lines.",
                            },
                        },
                        "required": ["find", "replace"],
                    },
                },
            },
            "required": ["path", "edits"],
            "$schema": "http://json-schema.org/draft-07/schema#",
        },
    },
}


@dataclass
class LineEdit:

    line_start: int
    line_end: int
    replace: str
    applied: bool = False


def remove_empty_lines(s: str) -> str:
    while "\n\n" in s:
        s = s.replace("\n\n", "\n")
    return s


def fuzzy_match(edit_find: str, lines: list[str]) -> tuple[int, int] | None:
    if not lines:
        return None

    num_edit_find_lines = len(edit_find.split("\n"))
    edit_find_trimmed = remove_empty_lines(edit_find)
    first_edit_find_line = edit_find_trimmed.split("\n")[0].strip()
    edit_find_without_spaces_len = len("".join(edit_find_trimmed.split()))
    best_ratio = 0.0
    best_start = -1
    best_window_len = 1
    for i in range(len(lines)):
        if lines[i].strip() != first_edit_find_line:
            continue

        for j in range(1, min(num_edit_find_lines * 2 + 1, len(lines) - i)):
            window_text = remove_empty_lines("\n".join(lines[i : i + j]))
            window_text_without_spaces_len = len("".join(window_text.split()))

            if abs(edit_find_without_spaces_len - window_text_without_spaces_len) / (edit_find_without_spaces_len or 1) > 0.05:
                continue

            ratio = SequenceMatcher(None, edit_find_trimmed, window_text, autojunk=False).ratio()
            if ratio > best_ratio:
                best_ratio = ratio
                best_start = i
                best_window_len = j
                if best_ratio == 1.0:
                    return best_start + 1, best_start + best_window_len

    if best_ratio >= 0.9:
        return best_start + 1, best_start + best_window_len

    return None


def convert_edit_to_line_edit(edit: dict, original_content: str) -> LineEdit:
    if not original_content and not edit["find"]:
        return LineEdit(1, 1, edit["replace"], False)

    lines = original_content.split("\n")
    if not edit["find"]:
        return LineEdit(1, len(lines), edit["replace"], False)

    edit_lines = edit["find"].split("\n")

    # First try exact matches
    for i in range(len(lines)):
        if lines[i : i + len(edit_lines)] == edit_lines:
            return LineEdit(i + 1, i + len(edit_lines), edit["replace"], False)

    # Then try matching without spacing
    for i in range(len(lines)):
        stripped_lines = [line.strip() for line in lines[i : i + len(edit_lines)]]
        stripped_edit_lines = [line.strip() for line in edit_lines]
        if stripped_lines == stripped_edit_lines:
            return LineEdit(i + 1, i + len(edit_lines), edit["replace"], False)

    # Finally try sequence matching while ignoring whitespace
    fuzzy_edit = fuzzy_match(edit["find"], lines)
    if fuzzy_edit:
        return LineEdit(*fuzzy_edit, edit["replace"], False)

    raise Exception("Edit not found in file")


def edit_file(path: str, edits: list[dict], yolo_mode: bool = False) -> str:
    def find_line_edit(n: int) -> LineEdit | None:
        for line_edit in line_edits:
            if line_edit.line_start <= n <= line_edit.line_end:
                return line_edit

        return None

    file_exists = Path(path).exists()
    if file_exists:
        with open(path) as f:
            original_content = f.read()
    else:
        original_content = ""

    line_edits: list[LineEdit] = [
        convert_edit_to_line_edit(edit, original_content)
        for edit in edits
    ]

    final_lines = []
    original_lines = original_content.split("\n")
    for i, line in enumerate(original_lines):
        line_num = i + 1
        line_edit = find_line_edit(line_num)
        if not line_edit:
            final_lines.append(line)
            continue

        if line_edit.applied:
            continue

        replace_lines = line_edit.replace.split("\n")
        if line_edit.replace:
            final_lines.extend(replace_lines)
        original_lines = original_content.split("\n")
        line_edit.applied = True

    unified_diff_md = ""
    for line in unified_diff(
        [f"{line}\n" for line in original_lines],
        [f"{line}\n" for line in final_lines],
        fromfile=path,
        tofile=path,
    ):
        unified_diff_md += line

    console.print()
    console.print(Markdown(f"```diff\n{unified_diff_md}\n```"))

    if not yolo_mode:
        answer = console.input("\n[bold]Run?[/] ([bold]Y[/]/n): ").strip().lower()
        if answer not in ("yes", "y", ""):
            reason = console.input("Why not? (optional, press Enter to skip): ").strip()
            return f"User declined: {reason or 'no reason'}"
    console.print(" └ Running...")
    try:
        with open(path, "w") as f:
            f.write("\n".join(final_lines))
    except Exception as e:
        console.print("   [red]edit_file failed[/red]")
        console.print(f"   [red]{str(e).strip()}[/red]")
        return f"edit_file failed: {str(e).strip()}"

    console.print("   [green]Command succeeded[/green]")
    return f"Successfully edited {path}"



from src.constants import console

READ_FILE_TOOL = {
    "type": "function",
    "function": {
        "name": "read_file",
        "description": "Read a file's contents (the first 1000 lines by default). When reading a file for the first time, do not change the defaults and always read the first 1000 lines unless you are absolutely certain of which lines need to be read.",
        "parameters": {
            "type": "object",
            "properties": {
                "path": {
                    "type": "string",
                    "description": "Relative path of the file",
                },
                "start": {
                    "type": "integer",
                    "description": "Which line to start reading from",
                    "default": 1,
                },
                "num_lines": {
                    "type": "integer",
                    "description": "Number of lines to read, defaults to 1000",
                    "default": 1000,
                },
            },
            "required": ["path"],
            "$schema": "http://json-schema.org/draft-07/schema#",
        },
    },
}


def read_file(path: str, start: int = 1, num_lines: int = 1000, yolo_mode: bool = False) -> str:
    if start == 1 and num_lines == 1000:
        console.print(f" └ Reading file [bold]{path}[/]...")
    else:
        last_line = start + num_lines - 1
        console.print(f" └ Reading file [bold]{path}[/] (lines {start}-{last_line})...")

    try:
        with open(path) as f:
            lines = []
            line_num = 1
            for line in f:
                if line_num < start:
                    line_num += 1
                    continue

                lines.append(line)
                line_num += 1

                if len(lines) >= num_lines:
                    lines.append("... (truncated)")
                    break

            console.print("   [green]Command succeeded[/green]")
            return "\n".join(lines)
    except Exception as e:
        console.print("   [red]read_file failed[/red]")
        console.print(f"   [red]{str(e).strip()}[/red]")
        return f"read_file failed: {e}"



from prompt_toolkit.completion import (
    CompleteEvent,
    Completer,
    Completion,
)
from prompt_toolkit.document import Document
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from src.constants import CWD_PATH


class FileCompleter(Completer):

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> list[Completion]:
        """Return file completions for text after @ anywhere in input."""
        text = document.text
        cursor_pos = document.cursor_position
        text_before_cursor = document.text_before_cursor
        last_at_pos = text_before_cursor.rfind("@")
        if last_at_pos == -1:
            return []

        text_after_at = text[last_at_pos + 1 : cursor_pos]

        base_dir = CWD_PATH
        search_pattern = text_after_at
        prefix = ""

        if "/" in text_after_at:
            dir_part, file_part = text_after_at.rsplit("/", 1)
            base_dir = CWD_PATH / dir_part
            search_pattern = file_part
            prefix = dir_part + "/"

        if not base_dir.is_dir():
            return []

        try:
            dirs = []
            files = []
            for entry in sorted(base_dir.iterdir()):
                name = entry.name
                if search_pattern in name:
                    full_path = prefix + name
                    if entry.is_dir():
                        dirs.append(Completion(full_path, display=name, start_position=-len(text_after_at), style="Blue"))
                    else:
                        files.append(Completion(full_path, display=name, start_position=-len(text_after_at)))

            return dirs + files
        except (PermissionError, OSError):
            return []


def create_key_bindings() -> KeyBindings:
    bindings = KeyBindings()

    @bindings.add(Keys.Backspace)
    def backspace_with_completion(event):
        """Also trigger completion on backspace."""
        buffer = event.current_buffer
        buffer.delete_before_cursor(count=1)
        text_before_cursor = buffer.document.text_before_cursor
        last_at_pos = text_before_cursor.rfind("@")
        if last_at_pos != -1:
            buffer.start_completion()

    @bindings.add(Keys.ControlC)
    def ctrl_c_handler(event):
        """Ctrl+C should raise KeyboardInterrupt."""
        raise KeyboardInterrupt()

    return bindings



import os
from typing import Literal

import requests


class LLMClient:

    def __init__(self):
        self.host = os.environ.get("TASS_HOST", "http://localhost:8080")
        self.api_key = os.environ.get("TASS_API_KEY", "")

    def request(
        self,
        method: Literal["get", "post"],
        url: str,
        **kwargs,
    ):
        return requests.request(
            method,
            f"{self.host}{url}",
            headers={
                "Authorization": f"Bearer {self.api_key}",
            },
            **kwargs,
        )

    def get(self, url: str, **kwargs):
        return self.request("get", url, **kwargs)

    def post(self, url: str, **kwargs):
        return self.request("post", url, **kwargs)

    def get_models(self):
        return self.get("/v1/models", timeout=2)

    def get_chat_completions(self, messages: list[dict], tools: list[dict], stream: bool = False):
        return self.post(
            "/v1/chat/completions",
            json={
                "messages": messages,
                "tools": tools,
                "stream": stream,
                "chat_template_kwargs": {
                    "reasoning_effort": "medium",
                },
            },
            stream=stream,
        )



from .edit_file import EDIT_FILE_TOOL, edit_file
from .execute import EXECUTE_TOOL, execute
from .read_file import READ_FILE_TOOL, read_file

__all__ = [
    "EDIT_FILE_TOOL",
    "EXECUTE_TOOL",
    "READ_FILE_TOOL",
    "edit_file",
    "execute",
    "read_file",
]



import subprocess
from time import time

from rich.markdown import Markdown

from src.constants import (
    CWD_PATH,
    console,
)

READ_ONLY_COMMANDS = [
    "ls",
    "cat",
    "less",
    "more",
    "echo",
    "head",
    "tail",
    "wc",
    "grep",
    "find",
    "ack",
    "which",
    "sed",
    "find",
    "test",
]

EXECUTE_TOOL = {
    "type": "function",
    "function": {
        "name": "execute",
        "description": f"Executes a shell command. The current working directory for all commands will be {CWD_PATH}",
        "parameters": {
            "type": "object",
            "properties": {
                "command": {
                    "type": "string",
                    "description": "Full shell command to be executed.",
                },
                "explanation": {
                    "type": "string",
                    "description": "A brief explanation of why you want to run this command. Keep it to a single sentence.",
                },
            },
            "required": ["command", "explanation"],
            "$schema": "http://json-schema.org/draft-07/schema#",
        },
    },
}


def is_read_only_command(command: str) -> bool:
    """A simple check to see if the command is only for reading files.

    Not a comprehensive or foolproof check by any means, and will
    return false negatives to be safe.
    """
    if ">" in command:
        return False

    # Replace everything that potentially runs another command with a pipe
    command = command.replace("&&", "|")
    command = command.replace("||", "|")
    command = command.replace(";", "|")

    pipes = command.split("|")
    return all(pipe.strip().split()[0] in READ_ONLY_COMMANDS for pipe in pipes)


def execute(command: str, explanation: str, yolo_mode: bool = False) -> str:
    command = command.strip()
    requires_confirmation = not yolo_mode and not is_read_only_command(command)
    if requires_confirmation:
        console.print(Markdown(f"```shell\n{command}\n```"))
        if explanation:
            console.print(f"Explanation: {explanation}")
        answer = console.input("\n[bold]Run?[/] ([bold]Y[/]/n): ").strip().lower()
        if answer not in ("yes", "y", ""):
            reason = console.input("Why not? (optional, press Enter to skip): ").strip()
            return f"User declined: {reason or 'no reason'}"
        console.print(" └ Running...")
    else:
        console.print(f" └ Running [bold]{command}[/] (Explanation: {explanation})")

    try:
        start = time()
        result = subprocess.run(
            command,
            shell=True,
            capture_output=True,
            text=True,
        )
        took = time() - start
    except Exception as e:
        console.print("   [red]subprocess.run failed[/red]")
        console.print(f"   [red]{str(e).strip()}[/red]")
        return f"subprocess.run failed: {str(e).strip()}"

    out = result.stdout
    err = result.stderr.strip()
    if result.returncode == 0:
        console.print(f"   [green]Command succeeded[/green] [grey50]({took:,.2f}s)[/grey50]", highlight=False)
    else:
        console.print(f"   [red]Command failed[/red] (code {result.returncode}) [grey50]({took:,.2f}s)[/grey50]", highlight=False)
        if err:
            console.print(f"   [red]{err}[/red]")

    if len(out.split("\n")) > 1000:
        out_first_1000 = "\n".join(out.split("\n")[:1000])
        out = f"{out_first_1000}... (Truncated)"

    if len(err.split("\n")) > 1000:
        err_first_1000 = "\n".join(err.split("\n")[:1000])
        err = f"{err_first_1000}... (Truncated)"

    if len(out) > 20000:
        out = f"{out[:20000]}... (Truncated)"

    if len(err) > 20000:
        err = f"{err[:20000]}... (Truncated)"

    return f"Command output (exit {result.returncode}):\n{out}\n{err}\nTook {took:,.2f}s"



import os
import platform
import subprocess
from pathlib import Path

from rich.console import Console

console = Console()
CWD_PATH = Path.cwd().resolve()


def get_shell_info() -> str:
    shell_path = os.environ.get('SHELL', '')
    shell_name = shell_path.split('/')[-1] if shell_path else 'unknown'
    return shell_name


def get_git_info() -> str:
    try:
        git_info = []

        subprocess.run(['git', 'rev-parse', '--git-dir'], capture_output=True, check=True, cwd=CWD_PATH)
        result = subprocess.run(['git', 'branch', '--show-current'], capture_output=True, text=True, cwd=CWD_PATH)
        branch = result.stdout.strip()
        if branch:
            git_info.append(f"Branch: {branch}")

        result = subprocess.run(['git', 'status', '--porcelain'], capture_output=True, text=True, cwd=CWD_PATH)
        has_changes = bool(result.stdout.strip())
        git_info.append(f"Uncommitted changes: {'Yes' if has_changes else 'No'}")

        return '\n'.join(git_info)
    except Exception:
        return "Not a git repository"


def get_directory_listing() -> str:
    try:
        entries = []
        iterdir = CWD_PATH.iterdir()
        for count, entry in enumerate(sorted(iterdir)):
            if count >= 100:
                entries.append(f"... and {len(list(iterdir)) - 100} more items")
                break
            if entry.is_dir():
                entries.append(f"{entry.name}/")
            else:
                entries.append(entry.name)
        return '\n'.join(entries) if entries else "Empty directory"
    except PermissionError:
        return "Permission denied"


SYSTEM_PROMPT = f"""You are tass, or Terminal Assistant, a helpful AI that executes shell commands based on natural-language requests.

If the user's request involves making changes to the filesystem such as creating or deleting files or directories, you MUST first check whether the file or directory exists before proceeding.

If a user asks for an answer or explanation to something instead of requesting to run a command, answer briefly and concisely. Do not supply extra information, suggestions, tips, or anything of the sort.

This app has a feature where the user can refer to files or directories by typing @ which will open an file autocomplete dropdown. When this feature is used, the @ will remain in the filename. When working with said file, ignore the preceding @.

Current working directory: {CWD_PATH}

# Directory Context
OS: {platform.system()}
Shell: {get_shell_info()}

# Git Info
{get_git_info()}

# Directory Listing
{get_directory_listing()}
"""



import argparse

from src.app import TassApp


def main():
    parser = argparse.ArgumentParser(
        description="Terminal Assistant - Ask an LLM to run commands"
    )
    parser.add_argument(
        "--yolo",
        action="store_true",
        help="YOLO mode: execute all commands and edit files without asking for confirmation",
    )
    parser.add_argument(
        "prompt",
        nargs="?",
        help="Prompt to run (enclose in quotes; runs in single-shot mode and exits)",
    )
    args = parser.parse_args()

    app = TassApp(yolo_mode=args.yolo)

    if args.prompt:
        app.run(initial_input=args.prompt)
    else:
        app.run()



import json

from prompt_toolkit import prompt
from rich.console import Group
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.text import Text

from src.constants import (
    SYSTEM_PROMPT,
    console,
)
from src.llm_client import LLMClient
from src.tools import (
    EDIT_FILE_TOOL,
    EXECUTE_TOOL,
    READ_FILE_TOOL,
    edit_file,
    execute,
    read_file,
)
from src.utils import (
    FileCompleter,
    create_key_bindings,
)


class TassApp:

    def __init__(self, yolo_mode: bool = False):
        self.yolo_mode = yolo_mode
        self.messages: list[dict] = [{"role": "system", "content": SYSTEM_PROMPT}]
        self.llm_client = LLMClient()
        self.key_bindings = create_key_bindings()
        self.file_completer = FileCompleter()
        self.context_tokens = 0
        self.TOOLS_MAP = {
            "execute": execute,
            "read_file": read_file,
            "edit_file": edit_file,
        }

    def check_llm_host(self):
        try:
            response = self.llm_client.get_models()
            console.print("Terminal Assistant [green](LLM connection ✓)[/green]")
            if response.status_code == 200:
                return
        except Exception:
            console.print("Terminal Assistant [red](LLM connection ✗)[/red]")

        console.print("\n[red]Could not connect to LLM[/red]")
        console.print(f"If your LLM isn't running on {self.llm_client.host}, you can set the [bold]TASS_HOST[/] environment variable to a different URL.")
        new_host = console.input(
            "Enter a different URL for this session (or press Enter to keep current): "
        ).strip()

        if new_host:
            self.llm_client.host = new_host

        try:
            response = self.llm_client.get_models()
            if response.status_code == 200:
                console.print(f"[green]Connection established to {self.llm_client.host}[/green]")
        except Exception:
            console.print(f"[red]Unable to verify new host {self.llm_client.host}. Continuing with it anyway.[/red]")

    def summarize(self):
        if self.context_tokens < 30_000 and len(self.messages) <= 50:
            return

        prompt = (
            "The conversation is becoming long and might soon go beyond the "
            "context limit. Please provide a detailed summary of the conversation, "
            "preserving all important details. Make sure context is not lost so that "
            "the conversation can continue without needing to reclarify anything. "
            "You don't have to preserve entire contents of files that have been read "
            " or edited, they can be read again if necessary."
        )

        console.print("\n - Summarizing conversation...")
        response = self.llm_client.get_chat_completions(
            messages=[*self.messages, {"role": "user", "content": prompt}],
            tools=[
                EDIT_FILE_TOOL,
                EXECUTE_TOOL,
                READ_FILE_TOOL,
            ],  # For caching purposes
        )
        data = response.json()
        summary = data["choices"][0]["message"]["content"]
        self.messages = [self.messages[0], {"role": "assistant", "content": f"Summary of the conversation so far:\n{summary}"}]
        console.print("   [green]Summarization completed[/green]")

    def call_llm(self) -> bool:
        response = self.llm_client.get_chat_completions(
            messages=self.messages,
            tools=[
                EDIT_FILE_TOOL,
                EXECUTE_TOOL,
                READ_FILE_TOOL,
            ],
            stream=True,
        )

        content = ""
        reasoning_content = ""
        tool_calls_map = {}
        timings_str = ""

        def generate_layout():
            groups = []

            if reasoning_content:
                last_three_lines = "\n".join(reasoning_content.rstrip().split("\n")[-3:])
                groups.append(Text(""))
                groups.append(
                    Panel(
                        Text(
                            last_three_lines,
                            style="grey50",
                        ),
                        title="Thought process",
                        title_align="left",
                        subtitle=timings_str,
                        style="grey50",
                    )
                )

            if content:
                groups.append(Text(""))
                groups.append(Markdown(content.rstrip()))
                if not reasoning_content and timings_str:
                    groups.append(Text(timings_str, style="grey50"))

            return Group(*groups)

        with Live(generate_layout(), refresh_per_second=10) as live:
            for line in response.iter_lines():
                line = line.decode("utf-8")
                if not line.strip():
                    continue

                if line == "data: [DONE]":
                    continue

                chunk = json.loads(line.removeprefix("data:"))
                if all(k in chunk.get("timings", {}) for k in ["cache_n", "prompt_n", "prompt_per_second", "predicted_n", "predicted_per_second"]):
                    timings = chunk["timings"]
                    self.context_tokens = timings["cache_n"] + timings["prompt_n"] + timings["predicted_n"]
                    timings_str = (
                        f"Input: {timings['prompt_n']:,} tokens, {timings['prompt_per_second']:,.2f} tok/s | "
                        f"Output: {timings['predicted_n']:,} tokens, {timings['predicted_per_second']:,.2f} tok/s | "
                        f"Context: {self.context_tokens:,} tokens"
                    )

                if chunk["choices"][0]["finish_reason"]:
                    live.update(generate_layout())

                delta = chunk["choices"][0]["delta"]
                if not any(delta.get(key) for key in ["content", "reasoning_content", "tool_calls"]):
                    continue

                if delta.get("reasoning_content"):
                    reasoning_content += delta["reasoning_content"]
                    live.update(generate_layout())

                if delta.get("content"):
                    content += delta["content"]
                    live.update(generate_layout())

                for tool_call_delta in delta.get("tool_calls") or []:
                    index = tool_call_delta["index"]
                    if index not in tool_calls_map:
                        tool_calls_map[index] = (
                            {
                                "index": index,
                                "id": "",
                                "type": "",
                                "function": {
                                    "name": "",
                                    "arguments": "",
                                },
                            }
                        )

                    tool_call = tool_calls_map[index]
                    if tool_call_delta.get("id"):
                        tool_call["id"] += tool_call_delta["id"]
                    if tool_call_delta.get("type"):
                        tool_call["type"] += tool_call_delta["type"]
                    if tool_call_delta.get("function"):
                        function = tool_call_delta["function"]
                        if function.get("name"):
                            tool_call["function"]["name"] += function["name"]
                        if function.get("arguments"):
                            tool_call["function"]["arguments"] += function["arguments"]

        self.messages.append(
            {
                "role": "assistant",
                "content": content.strip(),
                "reasoning_content": reasoning_content.strip(),
                "tool_calls": list(tool_calls_map.values()) or [],
            }
        )

        if not tool_calls_map:
            return True

        try:
            for tool_call in tool_calls_map.values():
                tool = self.TOOLS_MAP[tool_call["function"]["name"]]
                tool_args = json.loads(tool_call["function"]["arguments"])
                tool_args["yolo_mode"] = self.yolo_mode
                result = tool(**tool_args)
                self.messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call["id"],
                        "name": tool_call["function"]["name"],
                        "content": result,
                    }
                )
            return False
        except Exception as e:
            self.messages.append({"role": "user", "content": f"Tool call failed: {e}"})
            console.print(f"   [red]Tool call failed: {str(e).strip()}[/red]")
            return self.call_llm()

    def run(self, initial_input: str | None = None):
        if initial_input:
            self.messages.append({"role": "user", "content": initial_input})
            while True:
                try:
                    finished = self.call_llm()
                except Exception as e:
                    console.print(f"Failed to call LLM: {e}")
                    break

                if finished:
                    return

        try:
            self.check_llm_host()
        except KeyboardInterrupt:
            console.print("\nBye!")
            return

        while True:
            console.print()
            try:
                input_lines = []
                while True:
                    input_line = prompt(
                        "> ",
                        completer=self.file_completer,
                        complete_while_typing=True,
                        key_bindings=self.key_bindings,
                    )
                    if not input_line or input_line[-1] != "\\":
                        input_lines.append(input_line)
                        break
                    input_lines.append(input_line[:-1])
                user_input = "\n".join(input_lines)
            except KeyboardInterrupt:
                console.print("\nBye!")
                break

            if not user_input:
                continue

            if user_input.lower().strip() == "exit":
                console.print("\nBye!")
                break

            self.messages.append({"role": "user", "content": user_input})
            while True:
                try:
                    finished = self.call_llm()
                except Exception as e:
                    console.print(f"Failed to call LLM: {e}")
                    break

                if finished:
                    self.summarize()
                    break



from src.tools.edit_file import fuzzy_match

file_contents = """def fib(n):
    '''Return the nth Fibonacci number (iterative).'''
    if n <= 0:
        return 0
    if n == 1:
        return 1
    a, b = 0, 1
    for _ in range(2, n + 1):
        a, b = b, a + b
    return b


if __name__ == "__main__":
    # Print first 10 Fibonacci numbers
    for i in range(11):
        print(f"fib({i}) = {fib(i)}")"""


def test_fuzzy():
    edit_find = "    return b\n\n\n\nif __name__ == \"__main__\":"
    lines = file_contents.split("\n")
    assert fuzzy_match(edit_find, lines) == (10, 13)

    edit_find = """def fib(n):

'''Return the nth Fibonacci number iterative.'''

    if n<= 0:

         return 0

    if n ==1:

      return 1

    a, b = 0, 1

    for _ in range(2, n + 1):

        a, b = b, a + b

    return b"""
    assert fuzzy_match(edit_find, lines) == (1, 10)
